module std.symtable;

import std.string;
import std.demangle;
import std.file;
import std.regexp;

private {
    extern (C) int strtol(char* ptr, char** eptr, int base);
}

enum SymtableType {
    NONE, BINUTILS_NM, DMC_MAP
}

class Symtable {
    class Sym {
        uint addr;
        char[] name;
        this(uint a, char[] n) {
            addr = a;
            name = n;
        }
        int opCmp(Object o) {
            return addr - (cast(Sym)o).addr;
        }
    }

    this(char[] filename, SymtableType st) {
        switch (st) {
        case SymtableType.BINUTILS_NM:
            char[] cont = cast(char[])std.file.read(filename);
            foreach (char[] line; std.string.split(cont, "\n")) {
                char[][] toks = std.string.split(line, " ");
                if (toks.length != 3) continue;
                uint addr = strtol(toStringz(toks[0]), null, 16);
                char[] name = demangle(toks[2]);
                syms ~= new Sym(addr, name);
            }
            syms.sort;
            break;
        case SymtableType.DMC_MAP:
            char[] cont = cast(char[])std.file.read(filename);
            RegExp reg =
                new RegExp(` [\dABCDEF]{4}:[\dABCDEF]{8}\s{7}(.*)`, ``);
            foreach (char[] line; std.string.split(cont, "\n")) {
                if (reg.test(line) == 0) continue;
                line = reg.replace(`$1`);
                char[][] toks = std.string.split(line);
                if (toks.length != 2) continue;
                uint addr = strtol(toStringz(toks[1]), null, 16);
                char[] name = demangle(toks[0]);
                syms ~= new Sym(addr, name);
            }
            syms.sort;
            break;
        case SymtableType.NONE:
            break;
        default:
            throw new Error("unknown symtable type");
        }
    }

    char[] lookupName(uint addr) {
        for (int i = 0; i < syms.length; i++) {
            if (syms[i].addr > addr) {
                if (i == 0) return null;
                else return syms[i-1].name;
            }
        }
        return null;
    }

    Sym[] getAllSymbol() { return syms; }

    static void setDefaultSymtable(Symtable st) {
        default_symtable = st;
    }
    static Symtable getDefaultSymtable() {
        return default_symtable;
    }

protected:
    this() {}

private:
    Sym[] syms;
    static Symtable default_symtable;
}
